﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using SmartX;
using System.Windows.Forms;

namespace TechNote78_TCPClient_CS
{
    class CSmartTCPClientWrapping
    {
        // 생성자 1 : 프로젝트에서 SmartConfigs를 사용하지 않는 경우
        public CSmartTCPClientWrapping(SmartTCPClient pSmartTCPClient)
        {
            m_pSmartTCPClient = pSmartTCPClient;
            m_bConnected = false;
            m_bConnectFirst = false;
            m_TimerConnectRetry = null;
            m_bConnecting = false;
            m_bSendData = false;
            m_bAckCompletion = false;
            m_ListReceiveData = new List<byte>();
            smartConfig = new SmartConfigs();
        }

        // 생성자 2 : 프로젝트에서 SmartConfigs를 사용하는 경우
        public CSmartTCPClientWrapping(SmartTCPClient pSmartTCPClient, SmartConfigs pSmartConfig)
        {
            m_pSmartTCPClient = pSmartTCPClient;
            m_bConnected = false;
            m_bConnectFirst = false;
            m_TimerConnectRetry = null;
            m_bConnecting = false;
            m_bSendData = false;
            m_bAckCompletion = false;
            m_ListReceiveData = new List<byte>();
            smartConfig = pSmartConfig;
        }

        // 네트워크 상태를 진단하기 위한 열거값
        public enum NetworkConnectStatus : int
        {
            Available = 0,              // 서버와 연결 가능한 상태
            NetworkDeviceProblems = 1,  // 네트워크 장비(공유기, 허브) 문제 발생. Ping 통신 실패
            ServerProblems = 2,         // 서버 확인. Ping 통신 실패
            ServerProcessProblems = 3,  // 서버 프로그램 확인. Ping 통신은 성공했으나, Connect 에 실패함
        }

        // 데이터 전송 상태 체크를 위한 열거값
        public enum SendStatus : int
        {
            Success = 0,                // 데이터 전송 후 서버로부터 응답을 받음
            ConnectionFailed = 1,       // 서버와 연결되지 않음
            SendFailed = 2,             // 데이터 전송에 실패 (Send관련 메서드에서 False리턴)
            AckFailed = 3,              // 데이터 전송에 성공했으나, 서버로부터 응답이 오지 않음.
        }

        // 데이터 수신 상태 체크를 위한 열거값
        public enum CommandStatus : int
        {
            Success = 0,                // 데이터 전송 후 서버로부터 응답을 받음
            ConnectionFailed = 1,       // 서버와 연결되지 않음
            SendFailed = 2,             // 데이터 전송에 실패 (Send관련 메서드에서 False리턴)
            AckFailed = 3,              // 데이터 전송에 성공했으나, 서버로부터 응답이 오지 않음.
            CommandInvalid = 4,         // 수신받은 Command가 잘못됨
            DataInvalid = 5,            // 수신받은 데이터가 잘못됨
        }

        // 연결 상태 체크를 위한 열거값
        public enum ConnectionStatus : int
        {
            Connected = 0,              // 서버와 연결됨
            DisConnected = 1,           // 서버와 연결이 해제됨
            Connecting = 2,             // 서버와 연결중인 상태
        }

        // 데이터 송수신 패킷 구조체
        public struct Data_Packet
        {
            public byte[] chCommand;                // 송수신한 Command를 저장
            public byte chChannel;                  // 송수신한 Channel을 저장
            public byte[] chData;                   // 송수신한 Data를 저장
            public CommandStatus chCommandStatus;   // 수신받은 데이터의 상태를 저장
        }

        #region ********************************* [ 주요 멤버 변수 ] **************************************

        private SmartTCPClient m_pSmartTCPClient;       // SmartTCPClient 사용을 위한 변수
        private SmartConfigs smartConfig;               // SmartConfigs 사용을 위한
        private SmartTimer m_TimerConnectRetry;         // 연결 재시도를 위한 타이머

        private int m_iConnectTimeoutInterval;          // 동기방식일때 최대 연결시간을 저장할 변수
        private bool m_bConnected;                      // 현재 연결 상태를 확인할 플래그
        private bool m_bConnecting;                     // 현재 연결 시도 중인지 확인할 플래그
        private bool m_bConnectFirst;                   // 최초 연결시 Connect() 메서드를 빨리 호출하기 위한 플래그
        private int m_iConnectRetryCount;               // 연결 재시도 횟수를 저장할 변수

        private bool m_bSynchronous_Connect;            // Connect의 동기/비동기 방식을 저장할 변수
        private bool m_bSynchronous_Receive;            // Receive의 동기/비동기 방식을 저장할 변수

        private int m_iConnect_MaxRetryCount;           // 연결 재시도 횟수를 저장할 변수
        private int m_iConnect_RetryInterval;           // 연결 재시도 간격을 저장할 변수

        private string m_strNetworkDeviceIPAddress;     // 공유기(허브)의 IP를 저장할 변수
        private string m_strServerIPAddress;            // 서버의 IP를 저장할 변수

        private bool m_bSendData;                       // 클라이언트에서 서버로 Request 했는지 확인할 플래그
        private bool m_bAckCompletion;                  // 서버에서 올바르게 Response 했는지 확인할 플래그
        private List<byte> m_ListReceiveData;           // 수신한 데이터를 저장할 List 

        #endregion

        #region ********************************* [ 상수 정의 ] *******************************************

        private const int SERVERCONNECT_TIMEOUTINTERVAL = 30000;    // 서버 상태 확인 시 사용할 TimeOut 값
        private const int SENDDATA_RETRYCOUNT = 3;                  // 데이터 전송 실패 시 재전송 할 횟수
        private const int ACKTIMEOUT = 5000;                        // Request 후 Response를 대기할 TimeOut 값

        #endregion

        #region ********************************* [ 기본 로직 ] *******************************************

        // Connect 재시도 타이머
        private void ConnectRetryTimer_Tick(object sender, EventArgs e)
        {
            m_TimerConnectRetry.Stop();

            // 처음 Connect 하는 경우 다음 Connect 시간을 설정
            if (m_bConnectFirst == true)
            {
                m_TimerConnectRetry.Interval = m_iConnect_RetryInterval;
                m_bConnectFirst = false;
            }

            // Connect 방식에 따른 처리(동기, 비동기)
            if (m_bSynchronous_Connect == true)
            {
                bool bRet;  // 연결상태를 저장할 변수

                // 연결 제한시간 설정값에 따른 처리
                if (m_iConnectTimeoutInterval == -1)
                {
                    // 연결 제한시간 없음
                    bRet = m_pSmartTCPClient.Connect();
                }
                else
                {
                    // 연결 제한시간 있음
                    bRet = m_pSmartTCPClient.Connect(m_iConnectTimeoutInterval);
                }

                // 연결 상태에 따른 처리
                if (bRet == true)
                {
                    // 연결 성공 시 플래그 설정
                    m_TimerConnectRetry.Stop();
                    m_bConnected = true;
                    m_bConnecting = false;
                    return;
                }
                else
                {
                    // 연결 실패 시 재시도 횟수 증가
                    m_iConnectRetryCount++;
                }

                // 재시도 횟수에 따른 처리
                if (m_iConnectRetryCount < m_iConnect_MaxRetryCount)
                {
                    m_TimerConnectRetry.Start();
                }
                else
                {
                    // 재시도 횟수 만료 시 연결 실패 처리
                    m_TimerConnectRetry.Stop();
                    ConnectRetryFail();
                }
            }
            else
            {
                // 비동기 방식
                m_pSmartTCPClient.Connect();
            }
        }

        // 연결방식이 비동기 방식인 경우 이벤트로 리턴값을 확인
        private void smartTCPClient_OnConnected(bool bConnected)
        {
            // 연결 상태에 따른 처리
            if (bConnected == true)
            {
                // 연결 성공 시 플래그처리
                m_TimerConnectRetry.Stop();
                m_bConnected = true;
                m_bConnecting = false;
            }
            else
            {
                // 연결 실패 시 처리
                m_bConnected = false;
                m_iConnectRetryCount++;

                if (m_iConnectRetryCount < m_iConnect_MaxRetryCount)
                {
                    m_TimerConnectRetry.Start();
                }
                else
                {
                    m_TimerConnectRetry.Stop();
                    ConnectRetryFail();
                }
            }
        }

        // 연결 실패한 경우 통신에러 원인을 확인
        private void ConnectRetryFail()
        {
            NetworkConnectStatus eNetworkStatus;
            // NetWork Check
            eNetworkStatus = Network_Check();

            // 네트워크가 사용불가능할 때 사용자가 처리하도록 한다.
            if (eNetworkStatus != NetworkConnectStatus.Available)
            {
                ConnectRetryFail_User(eNetworkStatus);
            }
        }

        // 통신에러 원인을 확인
        private NetworkConnectStatus Network_Check()
        {
            // 공유기(허브)연결 확인 : Ping 체크
            if (IPAddressAvailableCheck() == false)
            {
                return NetworkConnectStatus.NetworkDeviceProblems;
            }

            // 서버연결 확인 : Ping 체크
            if (ServerAvailableCheck() == false)
            {
                return NetworkConnectStatus.ServerProblems;
            }

            // 서버 프로그램 확인 : Connect 체크
            if (ServerProcessCheck(SERVERCONNECT_TIMEOUTINTERVAL) == false)
            {
                return NetworkConnectStatus.ServerProcessProblems;
            }

            // 모두 정상인 경우
            return NetworkConnectStatus.Available;
        }

        // 서버 프로그램 확인 : Connect 체크
        private bool ServerProcessCheck(int iServerConnectTimeout)
        {
            bool bRet;  // 서버 프로그램 상태를 저장

            // 비동기 방식인 경우 동기 방식으로 변경하여 테스트
            if (m_bSynchronous_Connect == false)
            {
                m_pSmartTCPClient.ConnectCallBackType = SmartTCPClient.CALLBACKTYPE.Sync;
            }

            // Connect 시도 후 결과를 저장
            bRet = m_pSmartTCPClient.Connect(iServerConnectTimeout);

            // 비동기 방식인 경우 변경한 연결방식을 재설정
            if (m_bSynchronous_Connect == false)
            {
                m_pSmartTCPClient.ConnectCallBackType = SmartTCPClient.CALLBACKTYPE.Async;
            }

            // Connect 결과가 성공인경우 서버를 닫고 값을 리턴
            if (bRet == true)
            {
                m_pSmartTCPClient.Close();
                return true;
            }

            // 실패한 경우 False 리턴
            return false;
        }

        // 공유기(허브) 연결 확인 : Ping 체크
        private bool IPAddressAvailableCheck()
        {
            // 공유기(허브) IP 설정이 안되있는 경우 False 리턴
            try
            {
                if (smartConfig.IPSettings.DeviceIP == "")
                {
                    return false;
                }
            }
            catch (InvalidCastException)
            {
                return false;
            }

            // Ping 정보 확인을 위한 변수 생성
            SmartX.CIPSetting.PINGRESULTINFO pingResultInfo = new SmartX.CIPSetting.PINGRESULTINFO();

            // 공유기(허브)로 Ping 통신
            if (smartConfig.IPSettings.PingIP(m_strNetworkDeviceIPAddress, ref pingResultInfo) == true)
            {
                // Ping 성공 시 true 리턴
                return true;
            }

            // Ping 실패 시 false 리턴
            return false;
        }

        // 서버 연결 확인 : Ping 통신(서버에서 Ping을 막은 경우 확인 불가)
        private bool ServerAvailableCheck()
        {
            // Ping 정보 확인을 위한 변수 생성
            SmartX.CIPSetting.PINGRESULTINFO pingResultInfo = new SmartX.CIPSetting.PINGRESULTINFO();
            // 서버로 Ping 통신
            if (smartConfig.IPSettings.PingIP(m_strServerIPAddress, ref pingResultInfo) == true)
            {
                // Ping 성공 시 true 리턴
                return true;
            }

            // Ping 실패 시 false 리턴
            return false;
        }

        // Connect 재시도
        private void ConnectReTry()
        {
            // 현재 연결이 되있는 경우 연결해제
            if (m_bConnected == true)
            {
                m_pSmartTCPClient.Close();
                m_bConnected = false;
            }
            // 값 재설정 후 timer 시작
            m_bConnectFirst = true;
            m_iConnectRetryCount = 0;
            m_bConnecting = true;
            m_TimerConnectRetry.Start();
        }

        // 데이터를 실제로 전송하는 메서드
        private bool SmartTCPClient_SendData(byte[] datas)
        {
            bool bRet;  // 전송상태를 저장

            // 데이터 재전송 횟수(상수)만큼 반복하여 데이터 전송
            for (int i = 0; i < SENDDATA_RETRYCOUNT; i++)
            {
                // 데이터 전송 후 결과를 저장
                bRet = m_pSmartTCPClient.SendByte(datas);

                // 데이터 전송 성공 시 처리
                if (bRet == true)
                {
                    // Request 송신 성공
                    m_bSendData = true;

                    return true;
                }

                System.Threading.Thread.Sleep(100);
            }

            // 데이터 전송 실패 처리
            return false;
        }

        // 데이터 수신시 발생하는 이벤트
        private void SmartTCPClient_OnReceiveHandler(byte[] datas)
        {
            // 수신데이터를 리스트에 추가
            m_ListReceiveData.AddRange(datas);

            // Request 요청에 의한 Response 수신인지 확인
            if (m_bSendData == false)
            {
                // 서버에서 임의로 송신한 데이터
                // 수신받은 데이터를 큐에 추가
                m_EventResponsePacket.Enqueue(Response_Parsing(m_ListReceiveData.ToArray(), false));
                // 사용자 정의 코드 실행
                OnReceiveHandler_User(m_ListReceiveData.ToArray());
            }
            else
            {
                // Request 요청에 의한 Repsonse 응답 수신
                // Request 플래그 초기화
                m_bSendData = false;
                // Response 응답을 받아 플래그 설정
                m_bAckCompletion = true;
            }
        }

        #endregion

        #region ********************************* [ 외부 인터페이스 ] *************************************

        // 연결 재시도 횟수를 설정
        public int ConnectMaxRetryCount
        {
            set { m_iConnect_MaxRetryCount = value; }
            get { return m_iConnect_MaxRetryCount; }
        }

        // 연결 재시도 간격을 설정
        public int ConnectRetryInterval
        {
            set { m_iConnect_RetryInterval = value; }
            get { return m_iConnect_RetryInterval; }
        }

        // 서버 연결 모드를 설정(동기, 비동기)
        public bool Synchronous_Connect
        {
            set { m_bSynchronous_Connect = value; }
            get { return m_bSynchronous_Connect; }
        }

        // 서버 수신 모드를 설정(동기, 비동기)
        public bool Synchronous_Receive
        {
            set { m_bSynchronous_Receive = value; }
            get { return m_bSynchronous_Receive; }
        }

        // 공유기(허브) IP를 설정
        public string NetworkDeviceIP
        {
            set { m_strNetworkDeviceIPAddress = value; }
            get { return m_strNetworkDeviceIPAddress; }
        }

        // 공유기(허브)의 연결상태를 체크
        public bool IsNetworkAvailable
        {
            get { return IPAddressAvailableCheck(); }
        }

        // 서버 IP를 설정
        public string ServerAddress
        {
            set { m_strServerIPAddress = value; }
            get { return m_strServerIPAddress; }
        }

        // 서버와 연결 상태를 확인
        public ConnectionStatus IsConnected
        {
            get
            {
                if (m_bConnected == true)
                {
                    return ConnectionStatus.Connected;
                }
                else if (m_bConnecting == true)
                {
                    return ConnectionStatus.Connecting;
                }
                else
                {
                    return ConnectionStatus.DisConnected;
                }
            }
        }

        // 서버와 연결 시도
        public void Connect(int iPort, int iTimeoutInterval)
        {
            // 현재 연결중이 아닐 경우 진행
            if (m_bConnecting == false)
            {
                m_bConnecting = true;
                // IP주소 및 Port 설정
                m_pSmartTCPClient.ServerIPAddress = m_strServerIPAddress;
                m_pSmartTCPClient.Port = iPort;
                // 동기 방식일 경우 최대 연결 대기시간을 설정
                m_iConnectTimeoutInterval = iTimeoutInterval;
                // 처음 Connect할 경우 빨리 Connect 시도하도록 설정
                m_bConnectFirst = true;
                m_iConnectRetryCount = 0;
                // - 수정
                m_pSmartTCPClient.ConnectCallBackType = (SmartTCPClient.CALLBACKTYPE)(Convert.ToInt32(m_bSynchronous_Connect));
                m_pSmartTCPClient.SendCallBackType = SmartTCPClient.CALLBACKTYPE.Sync;
                m_pSmartTCPClient.ReceiveCallBackType = (SmartTCPClient.CALLBACKTYPE)(Convert.ToInt32(m_bSynchronous_Receive));

                // 동기, 비동기 방식설정
                if (m_pSmartTCPClient.ConnectCallBackType == SmartTCPClient.CALLBACKTYPE.Async)
                {
                    // 비동기 방식일 경우 연결상태를 체크하기 위한 이벤트를 추가
                    m_pSmartTCPClient.OnConnected -= smartTCPClient_OnConnected;
                    m_pSmartTCPClient.OnConnected += new SmartTCPClient.ConnectHandler(smartTCPClient_OnConnected);
                }

                // 연결 재시도시 사용할 Timer 생성
                m_TimerConnectRetry = new SmartTimer();
                m_TimerConnectRetry.Interval = 10;
                m_TimerConnectRetry.Stop();
                m_TimerConnectRetry.Tick += new EventHandler(ConnectRetryTimer_Tick);
                m_TimerConnectRetry.Start();

                // 데이터 수신 이벤트 추가
                m_pSmartTCPClient.OnReceiveHandler -= SmartTCPClient_OnReceiveHandler;
                m_pSmartTCPClient.OnReceiveHandler += new SmartTCPClient.ReceiveHandler(SmartTCPClient_OnReceiveHandler);
            }
        }

        // 서버와 연결 해제
        public void Disconnect()
        {
            // m_TimerConnectRetry가 생성되었을 때만 실행
            if (m_TimerConnectRetry != null)
            {
                m_TimerConnectRetry.Stop();
            }
            // 서버와 연결된 경우에만 실행
            if (m_bConnected == true)
            {
                m_pSmartTCPClient.Close();

                m_bConnected = false;
            }
        }

        // 데이터 전송 메서드
        public SendStatus SendData(byte[] sendData)
        {
            // 서버와 연결되어있지 않다면 ConnectionFailed을 리턴
            if (m_bConnected == false)
            {
                return SendStatus.ConnectionFailed;
            }

            // Request 후 Response 체크를 위한 플래그 초기화
            m_bAckCompletion = false;

            bool bRet;  // 데이터 송신 결과를 저장할 플래그

            // 데이터 전송 후 결과를 저장
            bRet = SmartTCPClient_SendData(sendData);

            // 데이터 전송에 실패했다면 SendFailed를 리턴
            if (bRet == false)
            {
                return SendStatus.SendFailed;
            }

            // Response 대기시간으로 설정한 시간동안 Response를 대기
            for (int i = 0; i < ACKTIMEOUT; i++)
            {
                // Response 수신시 처리
                if (m_bAckCompletion == true)
                {
                    // 플래그 초기화
                    m_bAckCompletion = false;

                    // Response 데이터를 처리하는 사용자 함수
                    ACK_OnReceiveHandler_User(m_ListReceiveData.ToArray());

                    // Request후 Response 성공시 Success를 리턴
                    return SendStatus.Success;
                }
                System.Threading.Thread.Sleep(1);
                Application.DoEvents();
            }

            // Response 수신 실패를 리턴
            return SendStatus.AckFailed;
        }

        #endregion

        #region ********************************* [ 사용자 정의 영역 ] *********************************

        private const byte m_bSTX = 0x40;             // STX : @
        private const byte m_bETX = 0x3B;             // ETX : ;

        private Data_Packet m_NowRequestPacket;       // Request한 데이터를 저장할 구조체 변수
        private Queue<Data_Packet> m_EventResponsePacket = new Queue<Data_Packet>();    // 서버에서 임의로 송신한 데이터를 저장할 Queue

        // Connect 실패 시 사용자가 처리할 함수
        private void ConnectRetryFail_User(NetworkConnectStatus eNetworkStatus)
        {
            SmartX.SmartMessageBox.Size(500, 200);
            // 사용자에게 알림
            DialogResult ret = SmartX.SmartMessageBox.Show("네트워크 환경을 점검하시기 바랍니다.!!! \n네트워크 점검을 하시려면 예(Yes), \n연결시도를 중지하시려면 아니오(No)를 선택하시기 바립니다.",
                                               "네트워크 점검", MessageBoxButtons.YesNo, MessageBoxIcon.Question);

            // 버튼 선택에 따른 처리
            if (ret == DialogResult.Yes)
            {
                // 예 선택시 네트워크 상태에 따른 처리
                switch (eNetworkStatus)
                {
                    case NetworkConnectStatus.Available:
                        ConnectReTry();
                        break;
                    case NetworkConnectStatus.NetworkDeviceProblems:
                        SmartX.SmartMessageBox.Show("내부네트워크 문제 발생 - 공유기확인, 단선확인");
                        m_bConnecting = false;
                        break;
                    case NetworkConnectStatus.ServerProblems:
                        SmartX.SmartMessageBox.Show("서버연결 문제 발생 - 서버확인, 서버Ping차단확인, 단선확인");
                        m_bConnecting = false;
                        break;
                    case NetworkConnectStatus.ServerProcessProblems:
                        SmartX.SmartMessageBox.Show("서버프로그램 상태문제 발생 - 서버프로그램확인, 서버 운영체제및 프로그램 확인");
                        m_bConnecting = false;
                        break;
                }
            }
            else
            {
                // 아니오 선택 시 Connect 중단
                m_bConnecting = false;
            }
        }

        // < Request and Response Type >
        private void ACK_OnReceiveHandler_User(byte[] receiveData)
        {
            // Request에 따른 Response의 사용자 처리
        }

        // < Server to Client Event Type >
        private void OnReceiveHandler_User(byte[] receiveData)
        {
            // 서버에서 일방적으로 송신한 데이터의 처리
        }

        // 전송받은 데이터를 검증 후 파싱하는 메서드. 이벤트에서 태움. bAckRecv 인자로 Ack수신인지 확인
        private Data_Packet Response_Parsing(byte[] chData, bool bAckRecv)
        {
            // 수신 리스트를 클리어
            m_ListReceiveData.Clear();

            Data_Packet response = new Data_Packet();   // 수신 데이터를 저장할 구조체

            // STX, ETX 체크
            if ((chData[0] != m_bSTX) || (chData[chData.Length - 1] != m_bETX))
            {
                // 비정상적인 데이터
                response.chCommandStatus = CommandStatus.DataInvalid;
                return response;
            }

            // Request에 의한 Response인 경우
            if (bAckRecv == true)
            {
                // Request한 Command와 Response 한 Command를 비교
                if (Encoding.ASCII.GetString(m_NowRequestPacket.chCommand, 0, m_NowRequestPacket.chCommand.Length) != Encoding.ASCII.GetString(chData, 1, 3))
                {
                    // 비정상적인 Command
                    response.chCommandStatus = CommandStatus.CommandInvalid;
                    return response;
                }

                // Channel 값이 전송한 데이터와 같지 않다면 
                if (m_NowRequestPacket.chChannel != chData[4])
                {
                    // 잘못된 Channel 수신
                    response.chCommandStatus = CommandStatus.DataInvalid;
                    return response;
                }

                // Request 여부에 따른 체크
                if (m_NowRequestPacket.chData != null)
                {
                    // Request한 데이터가 있다면. (SetFrame 이라면)
                    for (int i = 0; i < m_NowRequestPacket.chData.Length; i++)
                    {
                        if (m_NowRequestPacket.chData[i] != chData[5 + i])
                        {
                            // 잘못된 데이터 수신
                            response.chCommandStatus = CommandStatus.DataInvalid;
                            return response;
                        }
                    }
                }
                else
                {
                    // Request한 데이터가 없다면. (GetFrame 이라면)
                    if (chData.Length - 3 - m_NowRequestPacket.chCommand.Length == 0)
                    {
                        // 수신한 데이터가 없는 경우 잘못된 데이터 수신
                        response.chCommandStatus = CommandStatus.DataInvalid;
                        return response;
                    }
                }
            }

            // 정상적인 값이라면 데이터 패킷에 값을 저장
            response.chCommand = new byte[3];
            System.Buffer.BlockCopy(chData, 1, response.chCommand, 0, response.chCommand.Length);

            // 수신받은 Ch를 저장
            response.chChannel = new byte();
            response.chChannel = chData[1 + response.chCommand.Length];

            // 데이터가 정상적으로 수신되어 수신 데이터를 구조체에 저장
            response.chData = new byte[chData.Length - response.chCommand.Length - 3];
            System.Buffer.BlockCopy(chData, 2 + response.chCommand.Length, response.chData, 0, response.chData.Length);

            // 데이터 수신 정상
            response.chCommandStatus = CommandStatus.Success;

            // 수신 데이터가 저장된 구조체를 리턴
            return response;
        }

        // 서버로 Request 데이터를 전송하는 메서드
        private SendStatus Request_Command(string strCommand, int iCh, byte[] chData)
        {
            // 전송 패킷 생성
            m_NowRequestPacket = new Data_Packet();

            // Command 저장
            m_NowRequestPacket.chCommand = new byte[strCommand.Length];
            m_NowRequestPacket.chCommand = Encoding.ASCII.GetBytes(strCommand);

            // Channel 저장
            m_NowRequestPacket.chChannel = new byte();
            m_NowRequestPacket.chChannel = (byte)Convert.ToChar(iCh.ToString());

            // Data 저장
            if (chData != null)
            {
                // Set Frame인 경우 데이터를 패킷에 저장
                m_NowRequestPacket.chData = new byte[chData.Length];
            }
            // Get Frame인 경우 Null 값을 패킷에 저장
            m_NowRequestPacket.chData = chData;

            // 전송할 데이터를 저장할 List
            List<byte> lstSendData = new List<byte>();

            // STX 추가
            lstSendData.Add(m_bSTX);
            // Command 추가
            lstSendData.AddRange(m_NowRequestPacket.chCommand);
            // Ch 추가
            lstSendData.Add(m_NowRequestPacket.chChannel);

            // Set Data 추가 : Request 프레임이 Get 일 경우 데이터가 없음(null)
            if (m_NowRequestPacket.chData != null)
            {
                // Set Frame일 경우 프레임에 데이터를 추가
                lstSendData.AddRange(m_NowRequestPacket.chData);
            }

            // ETX 추가
            lstSendData.Add(m_bETX);

            // 데이터를 전송후 전송 상태를 리턴
            return SendData(lstSendData.ToArray());
        }

        #endregion

        #region ********************************* [ 사용자 정의 외부 인터페이스 영역 ] *********************************
        // [ Protocol Structure definition ]
        //
        // < Request and Response Type >
        // * Read, Write and Return always set.
        // Type - Read : Request Send Command - Client to Server Communication : Get Server Data
        //             STX[1Byte] + CMD[3Byte] + Channel[1Byte] + ETX[1Byte]
        //
        // Type - Write : Request Send Command - Client to Server Communication : Set Server Data
        //             STX[1Byte] + CMD[3Byte] + Channel[1Byte] + DATA[4Byte] + ETX[1Byte]
        //
        // Type - Return : Response Acknowledge Response - Server to Client Communication
        //             STX[1Byte] + CMD[3Byte] + Channel[1Byte] + RETURN_DATA[4Byte] + ETX[1Byte]
        //
        // < Server to Client Event Type >
        // Type - Event : Event Response Command - Server to Client Communication
        //            STX[1Byte] + CMD[3Byte] + Channel[1Byte] + DATA[4Byte] + ETX[1Byte]
        //
        // STX : 0x40
        // ETX : 0x3B
        //
        // < Command Code Define : ASCII Code >
        // TMR : Read Temperature Data 
        //       0x54, 0x4D, 0x52
        // TMW : Write Temperature Data 
        //       0x54, 0x4D, 0x57
        // HMR : Read Humidity Data 
        //       0x48, 0x4D, 0x52
        // HMW : Write Humidity Data 
        //       0x48, 0x4D, 0x57
        // WTR : Read Water Level Data
        //       0x57, 0x54, 0x52
        // WTW : Write Water Level Data 
        //       0x57, 0x54, 0x57
        // DTR : Read Distance Data
        //       0x44, 0x54, 0x52
        // DTW : Write Distance Data 
        //       0x44, 0x54, 0x57
        //
        // < Channel Code Define : ASCII Code >
        // Channel 0 : 0x30
        // Channel 1 : 0x31
        // Channel 2 : 0x32
        // Channel 3 : 0x33
        //
        // < Data Field Define : Binary Code >
        // Float Data


        // Queue에 저장된 패킷의 수를 읽는다.
        public int GetServerEventReceiveQueueCount
        {
            get { return m_EventResponsePacket.Count; }
        }

        // Queue에 저장된 패킷을 가져온다.
        public Data_Packet GetServerEventReceiveQueue()
        {
            if (m_EventResponsePacket.Count == 0)
            {
                Data_Packet retNull = new Data_Packet();
                return retNull;
            }
            return m_EventResponsePacket.Dequeue();
        }

        // Request-Response 구조일 때 외부에서 이용하는 메서드 -> 동기 방식이므로 이 메서드 내부에서 리턴까지 처리함
        public Data_Packet Command_RequestResponse(string strCommand, int iCh, byte[] chData)
        {
            // 데이터 전송 후 전송 상태를 받아옴
            SendStatus eSendStatus = Request_Command(strCommand, iCh, chData);

            // Response 받은 데이터를 저장할 구조체
            Data_Packet response = new Data_Packet();

            // 전송 상태에 따른 수신 상태 처리
            switch (eSendStatus)
            {
                case SendStatus.ConnectionFailed:
                    // 서버와 연결이 안된경우
                    response.chCommandStatus = CommandStatus.ConnectionFailed;
                    break;
                case SendStatus.SendFailed:
                    // 데이터 전송에 실패한 경우(메서드가 False 리턴)
                    response.chCommandStatus = CommandStatus.SendFailed;
                    break;
                case SendStatus.AckFailed:
                    // 서버에서 응답하지 않은 경우
                    response.chCommandStatus = CommandStatus.AckFailed;
                    break;
                case SendStatus.Success:
                    // Request 이후 올바르게 Response 된 경우 수신받은 데이터를 변환하여 저장
                    response = Response_Parsing(m_ListReceiveData.ToArray(), true);
                    break;
            }
            return response;
        }

        #endregion
    }
}
