﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace TechNote78_TCPClient_CS
{
    public partial class MainForm : Form
    {
        public MainForm()
        {
            InitializeComponent();
        }

        #region ***************** [멤버변수] *********************************

        private CSmartTCPClientWrapping m_cSmartTCPClient;  // WrappingClass
        private bool m_bNetStatus = false;                  // 네트워크 상태를 체크할 플래그
        private int m_iCount = 0;                           // 송수신 횟수를 저장할 변수

        #endregion

        #region ***************** [이벤트] *********************************

        private void MainForm_Load(object sender, EventArgs e)
        {
            // 현재 Form을 MainForm으로 설정
            smartForm1.MainForm = this;

            // TCP Wrapping Class 인스턴싱
            m_cSmartTCPClient = new CSmartTCPClientWrapping(smartTCPClient1);
            // m_cSmartTCPClient = new CSmartTCPClientWrapping(smartTCPClient1, smartConfigs1);

            // 기본 설정
            m_cSmartTCPClient.Synchronous_Connect = false;      // Connect 비동기
            m_cSmartTCPClient.Synchronous_Receive = false;      // Receive 비동기
            m_cSmartTCPClient.ServerAddress = "192.168.0.100";  // 서버 IP
            m_cSmartTCPClient.NetworkDeviceIP = "192.168.0.1";  // 공유기(네트워크 장비) IP
            m_cSmartTCPClient.ConnectMaxRetryCount = 3;         // 연결 재시도 횟수
            m_cSmartTCPClient.ConnectRetryInterval = 5000;      // 연결 재시도 간격

            labServerIP.Text = m_cSmartTCPClient.ServerAddress;
            labDeviceIP.Text = m_cSmartTCPClient.NetworkDeviceIP;
            labConStatus.Text = "DisConnected";

            // 리스트 설정
            SmartX.SmartListBox.COLUMNSTYLE[] columnStyle = new SmartX.SmartListBox.COLUMNSTYLE[3];

            columnStyle[0].iColumnWidth = 20;
            columnStyle[0].eColumnAlign = SmartX.SmartListBox.COLUMNALIGNS.CENTER;
            columnStyle[0].iOffsetColumn = 0;

            columnStyle[1].iColumnWidth = 90;
            columnStyle[1].eColumnAlign = SmartX.SmartListBox.COLUMNALIGNS.CENTER;
            columnStyle[1].iOffsetColumn = 0;

            columnStyle[2].iColumnWidth = 90;
            columnStyle[2].eColumnAlign = SmartX.SmartListBox.COLUMNALIGNS.CENTER;
            columnStyle[2].iOffsetColumn = 0;

            lsbSendRecv.ColumnStyles = columnStyle;

            // Connect 체크를 위한 타이머 설정
            timerConCheck.Interval = 500;

            // 서버에서 임의로 보내는 데이터를 체크하기 위한 타이머 설정
            timerServerSend.Interval = 500;

            // 네트워크 상태 점검
            timerNetCheck.Interval = 300;
            timerNetCheck.Start();

            // 콤보박스 설정 - 수정
            cbConnect_Sync.ItemList.AddItem("Sync");
            cbConnect_Sync.ItemList.AddItem("Async");
            cbConnect_Sync.SelectItemIndex = 1;

            cbRecv_Sync.ItemList.AddItem("Sync");
            cbRecv_Sync.ItemList.AddItem("Async");
            cbRecv_Sync.SelectItemIndex = 1;
        }

        // 프로그램 종료 시 작업
        private void MainForm_Closing(object sender, CancelEventArgs e)
        {
            if (m_cSmartTCPClient.IsConnected == CSmartTCPClientWrapping.ConnectionStatus.Connected) { m_cSmartTCPClient.Disconnect(); }
            if (timerConCheck.IsStart == true) { timerConCheck.Stop(); }
            if (timerNetCheck.IsStart == true) { timerNetCheck.Stop(); }
            if (timerServerSend.IsStart == true) { timerServerSend.Stop(); }
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            Connect();
        }

        private void btnDisConnect_Click(object sender, EventArgs e)
        {
            DisConnect();
        }

        private void btnListClear_Click(object sender, EventArgs e)
        {
            lsbSendRecv.ClearAll();
            m_iCount = 0;
        }

        // 네트워크 상태를 검사
        private void timerNetCheck_Tick(object sender, EventArgs e)
        {
            timerNetCheck.Stop();
            // 네트워크 디바이스 (공유기, 허브)와 연결상태를 확인
            if (m_cSmartTCPClient.IsNetworkAvailable == true)
            {
                m_bNetStatus = true;
                labNetStatus.Text = "Enabled";
            }
            else
            {
                m_bNetStatus = false;
                labNetStatus.Text = "Disabled";
                timerNetCheck.Start();
            }
        }

        // Connect 시도 이후 연결완료를 체크하는 타이머
        private void timerConCheck_Tick(object sender, EventArgs e)
        {
            timerConCheck.Stop();

            if (m_cSmartTCPClient.IsConnected == CSmartTCPClientWrapping.ConnectionStatus.Connected)
            {
                // 서버와 연결이 됬다면 라벨에 표시
                labConStatus.Text = "Connected!!";
                // 서버에서 임의로 보내는 데이터를 체크시작
                timerServerSend.Start();
            }
            else
            {
                // 서버와 연결이 되지 않았을 때 체크
                if (m_cSmartTCPClient.IsConnected == CSmartTCPClientWrapping.ConnectionStatus.Connecting)
                {
                    // 현재 서버와 연결중이라면 라벨에 연결중 표시 후 연결 재확인
                    labConStatus.Text = "Now Connecting..";
                    timerConCheck.Start();
                }
                else
                {
                    // 서버와 연결중이 아니라면 연결 끊김처리
                    labConStatus.Text = "DisConnected";
                    // 네트워크 상태를 체크
                    timerNetCheck.Start();
                }
            }
        }

        // 서버에서 임의로 보내는 데이터를 체크
        private void timerServerSend_Tick(object sender, EventArgs e)
        {
            timerServerSend.Stop();

            // Queue에 있는 아이템 개수가 0개 이상일 때
            if (m_cSmartTCPClient.GetServerEventReceiveQueueCount > 0)
            {
                // 데이터를 저장하기 위한 구조체 생성
                CSmartTCPClientWrapping.Data_Packet ReceiveData = new CSmartTCPClientWrapping.Data_Packet();
                // 큐의 데이터를 가져온다
                ReceiveData = m_cSmartTCPClient.GetServerEventReceiveQueue();

                // 수신받은 데이터가 정상적인 경우 데이터를 저장
                if (ReceiveData.chCommandStatus == CSmartTCPClientWrapping.CommandStatus.Success)
                {
                    // 수신받은 데이터에서 Command와 Channel을 파싱한다.
                    string strCMD = Encoding.ASCII.GetString(ReceiveData.chCommand, 0, ReceiveData.chCommand.Length);
                    int iCh = Convert.ToInt32(((char)ReceiveData.chChannel).ToString());
                    // 수신받은 Command와 Channel에 따라 타겟 라벨을 설정
                    SmartX.SmartLabel sLabel = new SmartX.SmartLabel();
                    switch (strCMD)
                    {
                        case "TMR":
                        case "TMW":
                            sLabel = (iCh == 0) ? labTMPCh0 : (iCh == 1) ? labTMPCh1 : (iCh == 2) ? labTMPCh2 : labTMPCh3;
                            break;
                        case "HMR":
                        case "HMW":
                            sLabel = (iCh == 0) ? labHUMCh0 : (iCh == 1) ? labHUMCh1 : (iCh == 2) ? labHUMCh2 : labHUMCh3;
                            break;
                        case "WTR":
                        case "WTW":
                            sLabel = (iCh == 0) ? labWTRCh0 : (iCh == 1) ? labWTRCh1 : (iCh == 2) ? labWTRCh2 : labWTRCh3;
                            break;
                        case "DTR":
                        case "DTW":
                            sLabel = (iCh == 0) ? labDTCCh0 : (iCh == 1) ? labDTCCh1 : (iCh == 2) ? labDTCCh2 : labDTCCh3;
                            break;
                        default:
                            ReceiveData.chCommandStatus = CSmartTCPClientWrapping.CommandStatus.CommandInvalid;
                            break;
                    }

                    // 수신받은 CommandStatus에 따라 리스트에 값 출력
                    ADD_Status_to_List(ReceiveData.chCommandStatus, false);

                    sLabel.Text = BitConverter.ToSingle(ReceiveData.chData, 0).ToString();
                }
            }
            timerServerSend.Start();
        }

        #endregion

        #region   ***************** [ Command에 따른 데이터 전송 ] *************

        // 버튼 클릭 시 Temp값을 읽어온다. : TMR
        private void btnTMPRead_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radTMPCh0.Checked == true) ? labTMPCh0 : (radTMPCh1.Checked == true) ? labTMPCh1 : (radTMPCh2.Checked == true) ? labTMPCh2 : labTMPCh3;
            int iCh = (radTMPCh0.Checked == true) ? 0 : (radTMPCh1.Checked == true) ? 1 : (radTMPCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 값을 리턴받는다.
            float fValue = GetCMDValue("TMR", iCh, out eSendStatus);
            // 수신 성공 시 라벨값을 변경
            if (eSendStatus == CSmartTCPClientWrapping.CommandStatus.Success)
            {
                sLabel.Text = fValue.ToString();
            }
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭시 서버의 TMP값을 변경한다. : TMW
        private void btnTMPSet_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radTMPCh0.Checked == true) ? labTMPCh0 : (radTMPCh1.Checked == true) ? labTMPCh1 : (radTMPCh2.Checked == true) ? labTMPCh2 : labTMPCh3;
            int iCh = (radTMPCh0.Checked == true) ? 0 : (radTMPCh1.Checked == true) ? 1 : (radTMPCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 송수신 상태를 리턴받는다.
            eSendStatus = SetCMDValue("TMW", iCh, sLabel.Text);
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭 시 HUM값을 읽어온다. : HMR
        private void btnHUMRead_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radHUMCh0.Checked == true) ? labHUMCh0 : (radHUMCh1.Checked == true) ? labHUMCh1 : (radHUMCh2.Checked == true) ? labHUMCh2 : labHUMCh3;
            int iCh = (radHUMCh0.Checked == true) ? 0 : (radHUMCh1.Checked == true) ? 1 : (radHUMCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 값을 리턴받는다.
            float fValue = GetCMDValue("HMR", iCh, out eSendStatus);
            // 수신 성공 시 라벨값을 변경
            if (eSendStatus == CSmartTCPClientWrapping.CommandStatus.Success)
            {
                sLabel.Text = fValue.ToString();
            }
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭시 서버의 HUM값을 변경한다. : HMW
        private void btnHUMSet_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radHUMCh0.Checked == true) ? labHUMCh0 : (radHUMCh1.Checked == true) ? labHUMCh1 : (radHUMCh2.Checked == true) ? labHUMCh2 : labHUMCh3;
            int iCh = (radHUMCh0.Checked == true) ? 0 : (radHUMCh1.Checked == true) ? 1 : (radHUMCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 송수신 상태를 리턴받는다.
            eSendStatus = SetCMDValue("HMW", iCh, sLabel.Text);
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭 시 WTR값을 읽어온다. : WTR
        private void btnWTRRead_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radWTRCh0.Checked == true) ? labWTRCh0 : (radWTRCh1.Checked == true) ? labWTRCh1 : (radWTRCh2.Checked == true) ? labWTRCh2 : labWTRCh3;
            int iCh = (radWTRCh0.Checked == true) ? 0 : (radWTRCh1.Checked == true) ? 1 : (radWTRCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 값을 리턴받는다.
            float fValue = GetCMDValue("WTR", iCh, out eSendStatus);
            // 수신 성공 시 라벨값을 변경
            if (eSendStatus == CSmartTCPClientWrapping.CommandStatus.Success)
            {
                sLabel.Text = fValue.ToString();
            }
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭시 서버의 WTR값을 변경한다. : WTW
        private void btnWTRSet_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radWTRCh0.Checked == true) ? labWTRCh0 : (radWTRCh1.Checked == true) ? labWTRCh1 : (radWTRCh2.Checked == true) ? labWTRCh2 : labWTRCh3;
            int iCh = (radWTRCh0.Checked == true) ? 0 : (radWTRCh1.Checked == true) ? 1 : (radWTRCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 송수신 상태를 리턴받는다.
            eSendStatus = SetCMDValue("WTW", iCh, sLabel.Text);
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭 시 DTC값을 읽어온다. : DTR
        private void btnDTCRead_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radDTCCh0.Checked == true) ? labDTCCh0 : (radDTCCh1.Checked == true) ? labDTCCh1 : (radDTCCh2.Checked == true) ? labDTCCh2 : labDTCCh3;
            int iCh = (radDTCCh0.Checked == true) ? 0 : (radDTCCh1.Checked == true) ? 1 : (radDTCCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 값을 리턴받는다.
            float fValue = GetCMDValue("DTR", iCh, out eSendStatus);
            // 수신 성공 시 라벨값을 변경
            if (eSendStatus == CSmartTCPClientWrapping.CommandStatus.Success)
            {
                sLabel.Text = fValue.ToString();
            }
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        // 버튼 클릭시 서버의 DTC값을 변경한다. : DTW
        private void btnDTCSet_Click(object sender, EventArgs e)
        {
            // 타겟 라벨과 채널을 설정
            SmartX.SmartLabel sLabel = (radDTCCh0.Checked == true) ? labDTCCh0 : (radDTCCh1.Checked == true) ? labDTCCh1 : (radDTCCh2.Checked == true) ? labDTCCh2 : labDTCCh3;
            int iCh = (radDTCCh0.Checked == true) ? 0 : (radDTCCh1.Checked == true) ? 1 : (radDTCCh2.Checked == true) ? 2 : 3;
            // 송수신 상태값을 저장할 변수
            CSmartTCPClientWrapping.CommandStatus eSendStatus = new CSmartTCPClientWrapping.CommandStatus();
            // 데이터 전송 후 송수신 상태를 리턴받는다.
            eSendStatus = SetCMDValue("DTW", iCh, sLabel.Text);
            // 송수신 상태를 리스트에 저장
            ADD_Status_to_List(eSendStatus, true);
        }

        #endregion

        #region ***************** [ 사용자 메서드 ] **************************

        // Connect 동작
        private void Connect()
        {
            // 네트워크 사용이 가능한 경우에만 Connect 시도
            if (m_bNetStatus == false) return;

            // 현재 서버와 연결이 안되있는 경우에만 Connect시도
            if (m_cSmartTCPClient.IsConnected == CSmartTCPClientWrapping.ConnectionStatus.DisConnected)
            {
                // 콤보박스 선택여부에 따라 연결방식을 변경.(SelectIndex = 0 : 비동기, 1 : 동기) - 수정
                m_cSmartTCPClient.Synchronous_Connect = Convert.ToBoolean(cbConnect_Sync.SelectItemIndex);
                m_cSmartTCPClient.Synchronous_Receive = Convert.ToBoolean(cbRecv_Sync.SelectItemIndex);

                // Connect 시도. Port : 1470 / ConnectTimeOut : 3sec(비동기 방식은 영향없음)
                m_cSmartTCPClient.Connect(1470, 3000);

                // Connect 시도 이후 연결 상태를 확인하기 위해 타이머를 시작
                timerConCheck.Start();
            }
        }

        // DisConnect 동작
        private void DisConnect()
        {
            if (timerConCheck.IsStart == true) { timerConCheck.Stop(); }
            if (timerServerSend.IsStart == true) { timerServerSend.Stop(); }
            m_cSmartTCPClient.Disconnect();
            labConStatus.Text = "DisConnected";
        }

        // CMD에 따른 값을 서버에서 가져온다. GET 프레임. 인자값 : Command, Channel, 상태값
        private float GetCMDValue(string strCMD, int iCh, out CSmartTCPClientWrapping.CommandStatus cStatus)
        {
            // 데이터 전송 후 서버로부터 응답받은 데이터를 저장한다.
            CSmartTCPClientWrapping.Data_Packet sDataPacket = m_cSmartTCPClient.Command_RequestResponse(strCMD, iCh, null);
            // 인자값에 수신받은 송수신 상태를 저장
            cStatus = sDataPacket.chCommandStatus;
            // 송수신 상태가 SUCCESS 인 경우만 동작
            if (sDataPacket.chCommandStatus == CSmartTCPClientWrapping.CommandStatus.Success)
            {
                // 값을 변환하여 리턴
                return BitConverter.ToSingle(sDataPacket.chData, 0);
            }
            // 송수신에 실패했다면 -1 리턴
            return -1;
        }

        // CMD에 따른 값을 서버에 설정한다. SET 프레임. 인자값 : Command, Channel, Send Data
        private CSmartTCPClientWrapping.CommandStatus SetCMDValue(string strCMD, int iCh, string strValue)
        {
            // 데이터 전송후 송수신 상태값을 리턴
            return m_cSmartTCPClient.Command_RequestResponse(strCMD, iCh, BitConverter.GetBytes(Convert.ToSingle(strValue))).chCommandStatus;
        }

        // 수신받은 데이터의 CommandStatus에 따라 리스트에 아이템 추가
        private void ADD_Status_to_List(CSmartTCPClientWrapping.CommandStatus eCommandStatus, bool bAck)
        {
            List<string> strRetString = new List<string>();

            strRetString.Clear();

            strRetString.Add((m_iCount++).ToString());

            // 송수신 카운트를 증가
            // string strRetString = (m_iCount++).ToString() + "&";
            // 송수신 상태에 따른 스트링 추가 - 수정
            switch (eCommandStatus)
            {
                case CSmartTCPClientWrapping.CommandStatus.ConnectionFailed:
                    // 서버와 연결이 되지 않음
                    strRetString.Add("연결확인");
                    break;
                case CSmartTCPClientWrapping.CommandStatus.SendFailed:
                    // 데이터 전송에 실패함
                    strRetString.Add("전송실패");
                    break;
                case CSmartTCPClientWrapping.CommandStatus.AckFailed:
                    // 데이터 전송은 성공했지만, 서버로부터 응답이 없음
                    strRetString.Add("전송성공");
                    strRetString.Add("응답없음");
                    break;
                case CSmartTCPClientWrapping.CommandStatus.CommandInvalid:
                    // 서버로부터 수신받은 데이터의 Command 가 잘못됨
                    if (bAck == true)
                    {
                        strRetString.Add("전송성공");
                        strRetString.Add("잘못된 Command");
                    }
                    else
                    {
                        strRetString.Add("서버전송");
                        strRetString.Add("잘못된 Command");
                    }
                    break;
                case CSmartTCPClientWrapping.CommandStatus.DataInvalid:
                    // 서버로부터 수신받은 데이터가 잘못됨
                    if (bAck == true)
                    {
                        strRetString.Add("전송성공");
                        strRetString.Add("잘못된 Data");
                    }
                    else
                    {
                        strRetString.Add("서버전송");
                        strRetString.Add("잘못된 Data");
                    }
                    break;
                case CSmartTCPClientWrapping.CommandStatus.Success:
                    // 데이터 전송후 서버로부터 응답받음
                    if (bAck == true)
                    {
                        strRetString.Add("전송성공");
                        strRetString.Add("수신성공");
                    }
                    else
                    {
                        strRetString.Add("서버전송");
                        strRetString.Add("수신성공");
                    }
                    break;
            }

            // 송수신 상태가 Success 가 아닌 경우 서버와의 연결 해제 후 연결상태를 확인
            if (eCommandStatus != CSmartTCPClientWrapping.CommandStatus.Success)
            {
                DisConnect();
                timerNetCheck.Start();
            }

            // 리스트에 아이템을 추가
            lsbSendRecv.AddItem(strRetString.ToArray());
        }

        #endregion
    }
}