﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace TechNote78_TCPMultiServer_CS
{
    public partial class MainForm : Form
    {
        public MainForm()
        {
            InitializeComponent();
        }

        // 송수신할 데이터를 저장할 구조체
        public struct Data_Packet
        {
            public byte[] chCommand;
            public byte chChannel;
            public byte[] chData;
        }

        #region ***************** [멤버변수] *********************************

        private int m_iCount = 0;             // 송수신 횟수를 체크할 변수
        private string m_strSendClient;       // 데이터를 전송할 Client
        private const byte m_bSTX = 0x40;     // STX : @
        private const byte m_bETX = 0x3B;     // ETX : ;

        #endregion

        #region ***************** [이벤트] ***********************************

        private void MainForm_Load(object sender, EventArgs e)
        {
            // 현재Form을 MainForm으로 설정합니다.
            smartForm1.MainForm = this;

            // Server 장비의 IP 설정
            if (smartConfigs1.IPSettings.DeviceIP != "192.168.0.100")
            {
                smartConfigs1.IPSettings.DHCPEnable = 0;
                smartConfigs1.IPSettings.DeviceIP = "192.168.0.100";
                smartConfigs1.IPSettings.SubNetMask = "255.255.255.0";
                smartConfigs1.IPSettings.GateWay = "192.168.0.1";
                smartConfigs1.IPSettings.Save();
                smartConfigs1.IPSettings.SetApply();

                SmartX.SmartMessageBox.Show("IP가 재설정되어 재부팅합니다.");
                SmartX.SmartConfigs.ReBoot();
            }

            // 리스트박스 설정
            SmartX.SmartListBox.COLUMNSTYLE[] columnStyle = new SmartX.SmartListBox.COLUMNSTYLE[3];

            columnStyle[0].eColumnAlign = SmartX.SmartListBox.COLUMNALIGNS.CENTER;
            columnStyle[0].iColumnWidth = 20;
            columnStyle[0].iOffsetColumn = 0;

            columnStyle[1].eColumnAlign = SmartX.SmartListBox.COLUMNALIGNS.CENTER;
            columnStyle[1].iColumnWidth = 90;
            columnStyle[1].iOffsetColumn = 0;

            columnStyle[2].eColumnAlign = SmartX.SmartListBox.COLUMNALIGNS.CENTER;
            columnStyle[2].iColumnWidth = 90;
            columnStyle[2].iOffsetColumn = 0;

            lsbRecvData.ColumnStyles = columnStyle;

            // 라벨설정
            labServerIP.Text = smartConfigs1.IPSettings.DeviceIP;
            labServerPort.Text = "1470";
            labServerStatus.Text = "Server Closed";
        }

        // 프로그램 종료 시 서버와 타이머를 종료
        private void MainForm_Closing(object sender, CancelEventArgs e)
        {
            if (smartTCPMultiServer1.IsStart == true) 
            { 
                smartTCPMultiServer1.Stop(); 
            }
        }

        // Server Open 클릭
        private void btnServerOpen_Click(object sender, EventArgs e)
        {
            Server_OPEN();
        }

        // Server Close 클릭
        private void btnServerClose_Click(object sender, EventArgs e)
        {
            Server_CLOSE();
        }

        // 리스트에서 선택한 Client의 연결을 해제
        private void btnCloseClient_Click(object sender, EventArgs e)
        {
            // 리스트에서 선택한 항목이 있는 경우
            if (m_strSendClient != "")
            {
                // 선택한 Client의 연결을 해제
                smartTCPMultiServer1.CloseClientIP(m_strSendClient);
                m_strSendClient = "";
            }
        }

        // 선택한 Client의 IP를 타겟으로 설정
        private void lsbClientList_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lsbClientList.SelectItemIndex != -1)
            {
                m_strSendClient = lsbClientList.Items[lsbClientList.SelectItemIndex];
            }
        }

        // Recv Data Status ListBox를 Clear
        private void btnClear_Click(object sender, EventArgs e)
        {
            lsbRecvData.ClearAll();
            m_iCount = 0;
        }

        // 서버에 클라이언트가 Connect/DisConnect 되면 발생하는 이벤트
        private void smartTCPMultiServer1_OnClientUpdate()
        {
            // 서버가 시작된 경우 True, 시작이 안된 경우 False
            if (smartTCPMultiServer1.IsStart == true)
            {
                // 연결 클라이언트의 리스트 초기화
                lsbClientList.ClearAll();

                if (smartTCPMultiServer1.ClientInfoList != null)
                {
                    // 배열로 저장된 ID와 IPAddress를 배열의 크기만큼 반복하면서
                    for (int i = 0; i < smartTCPMultiServer1.ClientInfoList.Length; i++)
                    {
                        // 리스트 박스에 클라이언트들을 추가
                        lsbClientList.AddItem(smartTCPMultiServer1.ClientInfoList[i].strIPAddress);
                    }
                }
                else
                {
                    lsbClientList.ClearAll();
                }
            }
        }

        // 클라이언트 연결시 타겟 설정
        private void smartTCPMultiServer1_OnClientAdd(SmartX.SmartTCPMultiServer.CLIENTSINFOS addClientInfo)
        {
            m_strSendClient = addClientInfo.strIPAddress;
        }

        // 데이터 수신 시 발생하는 이벤트
        private void smartTCPMultiServer1_OnReceiveHandler(SmartX.SmartTCPMultiServer.CHandleClient.READINFO datas)
        {
            // Response 할 Client 주소를 설정
            m_strSendClient = datas.strIPAddress;
            // Response 응답
            bool bSendStatus = Response_Parsing(datas.receiveDatas);
            // 수신받은 데이터의 유효성에 따라 리스트에 값을 추가
            lsbRecvData.AddItem((m_iCount++).ToString() + "&수신 성공" + ((bSendStatus == true) ? "&전송 성공" : "&전송 실패"));
        }

        #endregion

        #region ***************** [사용자 메서드] **************************

        // Server Open
        private void Server_OPEN()
        {
            // SmartTCPMultiSever 설정
            smartTCPMultiServer1.MaxReceiveBufferSize = 1000;
            smartTCPMultiServer1.MaxClient = 3;
            smartTCPMultiServer1.MaxReceiveBufferSize = 3000;
            smartTCPMultiServer1.Port = Int32.Parse(labServerPort.Text);
            smartTCPMultiServer1.ReceiveTimeout = 3000;
            smartTCPMultiServer1.SetMaxConnectionReturnMessage("Over Connection!!!", SmartX.SmartTCPMultiServer.RETURNMSGCODE.UNICODE);

            if (smartTCPMultiServer1.IsStart == false)
            {
                // 현재 서버가 오픈되지 않은 경우 서버를 시작
                smartTCPMultiServer1.Start();
            }
            if (smartTCPMultiServer1.IsStart == true)
            {
                // 서버가 오픈되면 라벨에 표시
                labServerStatus.Text = "Server Opened";
            }
            else
            {
                labServerStatus.Text = "Sever Closed";
            }
        }

        // Server Close
        private void Server_CLOSE()
        {
            if (smartTCPMultiServer1.IsStart == true)
            {
                // 현재 서버가 열려있다면 서버를 닫는다.
                smartTCPMultiServer1.Stop();
                if (smartTCPMultiServer1.IsStart == false)
                {
                    labServerStatus.Text = "Server Closed";
                }
            }
        }

        // 데이터 송신 메서드
        private bool Send_Data(string strCMD, int iCh, string strData)
        {
            List<byte> listSendData = new List<byte>();
            // STX 추가
            listSendData.Add(m_bSTX);
            // command 추가
            listSendData.AddRange(Encoding.ASCII.GetBytes(strCMD));
            // Channel 추가
            listSendData.AddRange(Encoding.ASCII.GetBytes(iCh.ToString()));
            // Data 추가
            listSendData.AddRange(BitConverter.GetBytes(Convert.ToSingle(strData)));
            // ETX 추가
            listSendData.Add(m_bETX);

            // 데이터 전송 결과를 리턴
            return smartTCPMultiServer1.SendByteByClientIP(m_strSendClient, listSendData.ToArray());
        }

        // 전송받은 데이터를 검증 후 파싱하는 메서드
        private bool Response_Parsing(byte[] chData)
        {
            // STX, ETX 체크
            if ((chData[0] != m_bSTX) || (chData[chData.Length - 1] != m_bETX))
            {
                return false;
            }
            // 수신받은 Command를 변환
            string strCMD = Encoding.ASCII.GetString(chData, 1, 3);
            // 채널값을 변환하여 저장하고 유효성을 검사
            int iCh = Convert.ToInt32(((char)chData[4]).ToString());
            if ((iCh < 0) && (iCh > 3))
            {
                return false;
            }

            // 데이터 타겟 라벨을 설정
            SmartX.SmartLabel labTarget = new SmartX.SmartLabel();
            switch (strCMD)
            {
                case "TMR":
                case "TMW":
                    labTarget = (iCh == 0) ? labTMPCh0 : (iCh == 1) ? labTMPCh1 : (iCh == 2) ? labTMPCh2 : labTMPCh3;
                    break;
                case "HMR":
                case "HMW":
                    labTarget = (iCh == 0) ? labHUMCh0 : (iCh == 1) ? labHUMCh1 : (iCh == 2) ? labHUMCh2 : labHUMCh3;
                    break;
                case "WTR":
                case "WTW":
                    labTarget = (iCh == 0) ? labWTRCh0 : (iCh == 1) ? labWTRCh1 : (iCh == 2) ? labWTRCh2 : labWTRCh3;
                    break;
                case "DTR":
                case "DTW":
                    labTarget = (iCh == 0) ? labDTCCh0 : (iCh == 1) ? labDTCCh1 : (iCh == 2) ? labDTCCh2 : labDTCCh3;
                    break;
                default:
                    // 수신받은 Command가 비정상인 경우 false 리턴
                    return false;
            }

            // Set Frame인 경우 라벨을 변경
            if (chData.Length > 6)
            {
                labTarget.Text = (BitConverter.ToSingle(chData, 5)).ToString();
            }

            // 데이터를 전송
            return Send_Data(strCMD, iCh, labTarget.Text);
        }

        #endregion

        #region   ***************** [ Command에 따른 데이터 전송 ] *************

        // TMR 데이터 전송
        private void btnTMPSend_Click(object sender, EventArgs e)
        {
            int iCh = (radTMPCh0.Checked == true) ? 0 : (radTMPCh1.Checked == true) ? 1 : (radTMPCh2.Checked == true) ? 2 : 3;
            string strSendData = (radTMPCh0.Checked == true) ? labTMPCh0.Text : (radTMPCh1.Checked == true) ? labTMPCh1.Text : (radTMPCh2.Checked == true) ? labTMPCh2.Text : labTMPCh3.Text;

            // 데이터 전송 후 전송 상태에 따라 리스트에 값 추가
            bool bSendStatus = Send_Data("TMR", iCh, strSendData);
            lsbRecvData.AddItem((m_iCount++).ToString() + ((bSendStatus == true) ? "&전송 성공&" : "&전송 실패&"));
        }

        // HUM 데이터 전송
        private void btnHUMSend_Click(object sender, EventArgs e)
        {
            int iCh = (radHUMCh0.Checked == true) ? 0 : (radHUMCh1.Checked == true) ? 1 : (radHUMCh2.Checked == true) ? 2 : 3;
            string strSendData = (radHUMCh0.Checked == true) ? labHUMCh0.Text : (radHUMCh1.Checked == true) ? labHUMCh1.Text : (radHUMCh2.Checked == true) ? labHUMCh2.Text : labHUMCh3.Text;

            // 데이터 전송 후 전송 상태에 따라 리스트에 값 추가
            bool bSendStatus = Send_Data("HMR", iCh, strSendData);
            lsbRecvData.AddItem((m_iCount++).ToString() + ((bSendStatus == true) ? "&전송 성공&" : "&전송 실패&"));
        }

        // WTR 데이터 전송
        private void btnWTRSend_Click(object sender, EventArgs e)
        {
            int iCh = (radWTRCh0.Checked == true) ? 0 : (radWTRCh1.Checked == true) ? 1 : (radWTRCh2.Checked == true) ? 2 : 3;
            string strSendData = (radWTRCh0.Checked == true) ? labWTRCh0.Text : (radWTRCh1.Checked == true) ? labWTRCh1.Text : (radWTRCh2.Checked == true) ? labWTRCh2.Text : labWTRCh3.Text;

            // 데이터 전송 후 전송 상태에 따라 리스트에 값 추가
            bool bSendStatus = Send_Data("WTR", iCh, strSendData);
            lsbRecvData.AddItem((m_iCount++).ToString() + ((bSendStatus == true) ? "&전송 성공&" : "&전송 실패&"));
        }

        // DTC 데이터 전송
        private void btnDTCSend_Click(object sender, EventArgs e)
        {
            int iCh = (radDTCCh0.Checked == true) ? 0 : (radDTCCh1.Checked == true) ? 1 : (radDTCCh2.Checked == true) ? 2 : 3;
            string strSendData = (radDTCCh0.Checked == true) ? labDTCCh0.Text : (radDTCCh1.Checked == true) ? labDTCCh1.Text : (radDTCCh2.Checked == true) ? labDTCCh2.Text : labDTCCh3.Text;

            // 데이터 전송 후 전송 상태에 따라 리스트에 값 추가
            bool bSendStatus = Send_Data("DTR", iCh, strSendData);
            lsbRecvData.AddItem((m_iCount++).ToString() + ((bSendStatus == true) ? "&전송 성공&" : "&전송 실패&"));
        }

        #endregion
    }
}